# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

r"""XRootD file storage support for Invenio.

This modules provide an Invenio file storage interface for XRootD. By default
Invenio already has support for XRootD via
`PyFilesytem <http://docs.pyfilesystem.org/en/latest/>`_  and the
`XRootDPyFS <http://xrootdpyfs.readthedocs.io/>`_ package. This module
adds optimization and performance improvements such as efficient checksum
calculations.

Using this module is as simple as configuring your Invenio instance to
use the storage factory:

.. code-block:: python

   # config.py
   FILES_REST_STORAGE_FACTORY = \
       'invenio_xrootd:xrootd_storage_factory'

The module also provides a file storage interface for CERN EOS disk storage
system via XRootD protocol.

Using the EOS file storage module is as easy as configuring:

.. code-block:: python

   # config.py
   FILES_REST_STORAGE_FACTORY = \
       'invenio_xrootd:eos_storage_factory'

Overwriting reported checksum
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
XRootD v3.x only has support for reporting adler32 as the used checksum
algorithm, even though the server might be providing e.g. an MD5 checksum. If
this is the case, you can overwrite the reported checksum using the
configuration variable ``XROOTD_CHECKSUM_ALGO``:

.. code-block:: python

   # config.py
   XROOTD_CHECKSUM_ALGO = 'md5'

Kerberos authentication
~~~~~~~~~~~~~~~~~~~~~~~
If your XRootD server requires Kerberos authentication (e.g. CERN EOS), then
you can run a tool such as
`k5start <https://www.eyrie.org/~eagle/software/kstart/k5start.html>`_ on each
client node in order to obtain a Kerberos ticket and continue keeping the
ticket valid. The XRootD python bindings will transparently use this Kerberos
ticket to authenticate against your server.
"""

from __future__ import absolute_import, print_function

from .errors import SizeRequiredError
from .storage import EOSFileStorage, XRootDFileStorage, eos_storage_factory, \
    xrootd_storage_factory
from .version import __version__

__all__ = (
    '__version__',
    'eos_storage_factory',
    'EOSFileStorage',
    'SizeRequiredError',
    'xrootd_storage_factory',
    'XRootDFileStorage',
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Storage tests."""

from __future__ import absolute_import, print_function

from os.path import exists

import pytest

from invenio_xrootd import EOSFileStorage, SizeRequiredError, \
    XRootDFileStorage, eos_storage_factory, xrootd_storage_factory


def test_version():
    """Test version import."""
    from invenio_xrootd import __version__
    assert __version__


def test_init(file_url):
    """Test checksum overwrite."""
    assert XRootDFileStorage(file_url).checksum_algo is None
    assert EOSFileStorage(file_url).checksum_algo is None


def test_init_app(app, file_url):
    """Test checksum overwrite."""
    assert XRootDFileStorage(file_url).checksum_algo == 'md5'
    assert EOSFileStorage(file_url).checksum_algo == 'md5'


def test_factory(file_instance_mock):
    """Test factory functions."""
    assert isinstance(
        xrootd_storage_factory(fileinstance=file_instance_mock),
        XRootDFileStorage
    )
    assert isinstance(
        eos_storage_factory(fileinstance=file_instance_mock),
        EOSFileStorage
    )


def test_checksum_fallback(xrd_storage, eos_storage, file_md5):
    """Test checksum."""
    assert xrd_storage.checksum() == 'md5:{0}'.format(file_md5)
    assert eos_storage.checksum() == 'md5:{0}'.format(file_md5)


def test_checksum_xrd(xrd_storage_mocked, file_md5):
    """Test checksum."""
    assert xrd_storage_mocked.checksum() == 'adler32:{0}'.format(file_md5)


def test_checksum_xrd_overwrite(app, xrd_storage_mocked, file_md5):
    """Test checksum."""
    assert xrd_storage_mocked.checksum() == 'md5:{0}'.format(file_md5)


def test_checksum_ospath(xrd_storage_ospath, file_md5):
    """Test checksum overwrite."""
    assert xrd_storage_ospath.checksum() == 'md5:{0}'.format(file_md5)


def test_get_fs(xrd_storage, file_path, file_url):
    """Test checksum overwrite."""
    assert not exists(file_path)
    xrd_storage.initialize(10)
    assert exists(file_path)


def test_get_fs_ospath(xrd_storage_ospath, file_path):
    """Test checksum overwrite."""
    assert not exists(file_path)
    xrd_storage_ospath.initialize(10)
    assert exists(file_path)


def test_eos_initialize(eos_storage, file_path, file_url):
    """Test checksum overwrite."""
    assert not exists(file_path)
    eos_storage.initialize(size=10)
    assert exists(file_path)
    # Check parameter passing
    eos_storage.initialize(10)
    # Size required.
    pytest.raises(
        SizeRequiredError,
        eos_storage.initialize
    )


def test_eos_save(eos_storage, file_path, file_url, BytesIO):
    """Test checksum overwrite."""
    assert not exists(file_path)
    pytest.raises(
        SizeRequiredError,
        eos_storage.save,
        BytesIO(b'test')
    )
    assert not exists(file_path)
    eos_storage.save(BytesIO(b'test'), size=4)
    assert exists(file_path)


def test_eos_default_bookingsize(app, eos_storage, file_path, file_url,
                                 BytesIO):
    """Test checksum overwrite."""
    max_len = app.config['MAX_CONTENT_LENGTH']
    assert eos_storage.default_booking_size == max_len
    assert eos_storage.save(BytesIO(b'test'))
    assert exists(file_path)
    eos_storage.save(BytesIO(b'a' * (max_len + 1)), size=max_len + 1)
    assert exists(file_path)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Pytest configuration."""

from __future__ import absolute_import, print_function

import hashlib
import shutil
import tempfile
from functools import wraps
from os import makedirs
from os.path import dirname, join

import pytest
from flask import Flask

from invenio_xrootd import EOSFileStorage, XRootDFileStorage


@pytest.fixture
def BytesIO():
    """IO instance."""
    try:
        from io import BytesIO
        return BytesIO
    except ImportError:
        from StringIO import StringIO
        return StringIO


@pytest.yield_fixture()
def app():
    """Flask application fixture."""
    app = Flask('testapp')
    app.config.update(
        TESTING=True,
        MAX_CONTENT_LENGTH=10,
        XROOTD_CHECKSUM_ALGO='md5'
    )
    with app.app_context():
        yield app


@pytest.fixture
def mkurl():
    """Generate test root URL."""
    return lambda p: "root://localhost/{0}".format(p)


@pytest.yield_fixture
def tmppath():
    """Temporary path."""
    path = tempfile.mkdtemp()
    # shutil.copytree(join(dirname(__file__), "data"), join(path, "data"))
    yield path
    shutil.rmtree(path)


@pytest.fixture
def file_path(tmppath):
    """File path."""
    return join(tmppath, 'a/b/testfile')


@pytest.fixture
def file_url(file_path, mkurl):
    """File URL."""
    return mkurl(file_path)


@pytest.fixture
def file_content(file_path):
    """File content."""
    makedirs(dirname(file_path))
    data = b'test'
    with open(file_path, 'wb') as fp:
        fp.write(data)
    return data


@pytest.fixture
def file_md5(file_content):
    """Checksum of content."""
    m = hashlib.md5()
    m.update(file_content)
    return str(m.hexdigest())


@pytest.fixture
def xrd_storage(file_url):
    """Storage instance for XRootDFileStorage."""
    return XRootDFileStorage(file_url)


@pytest.fixture
def xrd_storage_ospath(file_path):
    """Storage instance for XRootDFileStorage with OS path."""
    return XRootDFileStorage(file_path)


@pytest.fixture
def xrd_storage_mocked(file_url, file_md5):
    """Patch xrd_checksum."""
    # Since local xrootd server can't compute checksums.
    xrd_storage = XRootDFileStorage(file_url)
    f = xrd_storage._get_fs

    @wraps(f)
    def mock(*args, **kwargs):
        fs, filename = f(*args, **kwargs)

        def xrd_checksum(*args, **kwargs):
            return 'adler32', file_md5

        fs.xrd_checksum = xrd_checksum
        return fs, filename

    xrd_storage._get_fs = mock

    return xrd_storage


@pytest.fixture
def eos_storage(file_url):
    """EOS storage instance."""
    return EOSFileStorage(file_url)


@pytest.fixture
def file_instance_mock(file_url):
    """Mock of a file instance."""
    class FileInstance(object):
        def __init__(self, **kwargs):
            for k, v in kwargs.items():
                setattr(self, k, v)

    return FileInstance(uri=file_url, size=4, updated=None)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""XRootD file storage interface."""

from __future__ import absolute_import, print_function

from functools import wraps

from flask import current_app
from fs.errors import UnsupportedError
from fs.path import basename, dirname
from invenio_files_rest.storage.pyfs import PyFSFileStorage, \
    pyfs_storage_factory
from xrootdpyfs import XRootDPyFS

from .errors import SizeRequiredError


class XRootDFileStorage(PyFSFileStorage):
    """File system storage using XRootD for accessing files.

    XRootD v3.x are only capable of reporting adler32 checksums, even though
    that the storage system will report e.g. a MD5 checksums. If this is the
    case, set the configuration variable ``XROOTD_CHECKSUM_ALOG`` to overwrite
    the alogrithm name reported by the XRootD server.
    """

    def __init__(self, fileurl, **kwargs):
        """Initialize file storage object."""
        # Overwrite reported checksum algorithm.
        self.checksum_algo = current_app.config.get(
            'XROOTD_CHECKSUM_ALGO') if current_app else None
        super(XRootDFileStorage, self).__init__(fileurl, **kwargs)

    def _get_fs(self, create_dir=True, query=None):
        """Get PyFilesystem instance and path."""
        # Fall-back to PyFS in case of non-xrootd URL
        if not self.fileurl.startswith(('root://', 'roots://')):
            return super(XRootDFileStorage, self)._get_fs(
                create_dir=create_dir)
        filedir = dirname(self.fileurl)
        filename = basename(self.fileurl)

        if query is None:
            fs = XRootDPyFS(filedir)
        else:
            fs = XRootDPyFS(filedir, query)

        if create_dir:
            fs.makedir('', recursive=True, allow_recreate=True)

        return (fs, filename)

    def checksum(self, chunk_size=None, progress_callback=None):
        """Compute checksum of file.

        Queries the XRootD server to get checksum if possible, otherwise falls
        back to default Python implementation. The checksum algorithm used
        will be the one configured on the XRootD server.
        """
        try:
            fs, path = self._get_fs()
            if not hasattr(fs, 'xrd_checksum'):
                raise UnsupportedError()
            algo, val = fs.xrd_checksum(path)
            return u'{0}:{1}'.format(self.checksum_algo or algo, val)
        except UnsupportedError:
            return super(XRootDFileStorage, self).checksum(
                chunk_size=chunk_size,
                progress_callback=progress_callback,
            )


def ensure_bookingsize(with_arg=False):
    """Decorator to set ``eos.bookingsize``."""
    def decorator(f):
        @wraps(f)
        def inner(self, *args, **kwargs):
            self.set_bookingsize(kwargs.get(
                'size',
                args[0] if args and with_arg else None
            ))
            res = f(self, *args, **kwargs)
            self.bookingsize = None
            return res
        return inner
    return decorator


class EOSFileStorage(XRootDFileStorage):
    """File storage for CERN EOS via XRootD.

    EOS needs to know the file size upfront in order to allocate your
    file to disk servers with enough available space. This is done by providing
    ``eos.bookingsize=xyz`` in the root URL.

    This file storage module takes care of setting the booking size for
    initialize and save operations.

    If you use this file storage class, it also means that your REST API
    clients **must** provide the file size upfront.
    """

    def __init__(self, fileurl, default_booking_size=None, **kwargs):
        """Initialize storage."""
        self.bookingsize = None
        self.default_booking_size = default_booking_size or \
            (current_app.config.get('MAX_CONTENT_LENGTH') if current_app
             else None)
        super(EOSFileStorage, self).__init__(fileurl, **kwargs)

    def set_bookingsize(self, size):
        """Set EOS booking size.

        Ensures that EOS will allocate the file to disk server with enough
        available space.
        """
        if size is None and self.default_booking_size is None:
            raise SizeRequiredError()
        self.bookingsize = size or self.default_booking_size

    def _get_fs(self, create_dir=True):
        """Get PyFilesystem instance and path."""
        query = {}
        if self.bookingsize:
            query['eos.bookingsize'] = self.bookingsize

        return super(EOSFileStorage, self)._get_fs(create_dir=create_dir,
                                                   query=query)

    @ensure_bookingsize(with_arg=True)
    def initialize(self, *args, **kwargs):
        """Initialize file of given size.

        Set the ``eos.bookingsize`` variable in the root URL, to ensure EOS
        allocates the file to a disk server with enough space.
        """
        return super(EOSFileStorage, self).initialize(*args, **kwargs)

    @ensure_bookingsize()
    def save(self, *args, **kwargs):
        """Initialize file of given size.

        Set the ``eos.bookingsize`` variable in the root URL, to ensure EOS
        allocates the file to a disk server with enough space.
        """
        return super(EOSFileStorage, self).save(*args, **kwargs)


def xrootd_storage_factory(**kwargs):
    """File storage factory for XRootD."""
    return pyfs_storage_factory(
        filestorage_class=XRootDFileStorage,
        **kwargs
    )


def eos_storage_factory(**kwargs):
    """File storage factory for EOS."""
    return pyfs_storage_factory(
        filestorage_class=EOSFileStorage,
        **kwargs
    )
